/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.basis;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Eine Formel fuer die Trefferpunkte einer Waffe.
 *
 * @author Bernhard Jung
 *
 */
public class Tp {

    /**
     * Prueft ob es sich bei einem String um eine gueltige TP-Formel handelt,
     * die von dieser Klasse verarbeitet werden kann.
     *
     * @param formel
     *            die ueberprueft werden soll.
     * @return {@code true}, wenn es sich um eine gueltige Formel handelt, sonst
     *         {@code false}.
     */
    public static final boolean istTpFormel( final String formel ) {

        final Pattern pattern =
            Pattern.compile( "([1-9]\\d*)?"
                           +  "."
                           + "([1-9]\\d*)?"
                           + "([+-]\\d+)?" );

        final Matcher matcher = pattern.matcher( formel );

        return matcher.matches();
    }

    private int anzahlWuerfel;

    private Wuerfel wuerfel;

    private int konstante;

    private final String formel;

    /**
     * @param dieFormel die Formel fuer die TP, z.B. "1W6+4"
     * @throws IllegalArgumentException
     *            wenn es sich nicht um eine gueltige TP-Formel handelt.
     */
    public Tp( final String dieFormel ) {

        if ( ! istTpFormel( dieFormel ) ) {

            throw new IllegalArgumentException(
                    dieFormel + " ist keine gueltige TP-Formel.");
        }

        dekodiereFormel( dieFormel );

        formel = dieFormel;
    }

    private void dekodiereFormel( final String formel ) {

        final int length = formel.length();

        int i = 0;

        if ( Character.isDigit( formel.charAt( 0 ) ) ) {

            final ErmittelteZahl zahl = ermitteleZahl( formel, i );

            i = zahl.offset;
            anzahlWuerfel = zahl.zahl;

        } else {

            anzahlWuerfel = 1;
        }

        // Den Buchstaben fuer den Wuerfel uebergehen.
        ++i;

        if ( ( i < length ) && Character.isDigit( formel.charAt( i ) ) ) {

            final ErmittelteZahl zahl = ermitteleZahl( formel, i );

            wuerfel = new Wuerfel( zahl.zahl );

            i = zahl.offset;

        } else {

            wuerfel = Wuerfel.W6;
        }

        if ( i >= length ) {

            // Die Formel ist zu Ende. Es gibt nur Wuerfelwuerfe und keine
            // Konstante zu aufaddieren.
            konstante = 0;

            return;
        }

        int faktor;

        if ( formel.charAt( i ) == '-' ) {

            faktor = -1;

            ++i;

        } else if ( formel.charAt( i ) == '+' ) {

            faktor = 1;

            ++i;

        } else {

            faktor = 1;
        }

        final ErmittelteZahl zahl = ermitteleZahl( formel, i );

        konstante = faktor * zahl.zahl;
    }

    /**
     * Ermittelt eine Zahl aus einem String.
     *
     * @param s
     *            der String in dem die Zahl steht.
     * @param offset
     *            ab dem die Zahl eingelesen werden soll. An der Position muss
     *            eine Zahl im String stehen. Es werden von der Position aus
     *            alle Zeichen eingelesen die Ziffern sind, um die Zahl zu
     *            bilden.
     * @return die im String stehende Zahl.
     */
    private ErmittelteZahl ermitteleZahl( final String s, final int offset ) {

        final int length = s.length();

        final StringBuilder builder = new StringBuilder();

        int i = offset;

        for ( ; i < length && Character.isDigit( s.charAt( i ) ) ; ++i ) {

            builder.append( s.charAt( i ) );
        }

        return new ErmittelteZahl( Integer.valueOf( builder.toString() ), i );
    }

    /**
     *
     * @param tpPlus
     *            Die Veraenderung der Trefferpunkte aus Koerperkraft.
     * @return die TP-Formel.
     */
    public final String getTpFormel( final int tpPlus ) {

        final int plus = konstante + tpPlus;

        if ( 0 != plus ) {

            return String.format( "%1$s%2$s%3$+d",
                                  anzahlWuerfel,
                                  wuerfel,
                                  plus );

        } else {

            return anzahlWuerfel + wuerfel.toString();
        }
    }

    @Override
    public String toString() {
        return formel;
    }

    private static class ErmittelteZahl {

        public int zahl;

        public int offset;

        ErmittelteZahl( final int zahl, final int offset ) {
            this.zahl = zahl;
            this.offset = offset;
        }
    }
}
