/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.basis.namen;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.gui.Util;
import de.bernhardjung.dsaprogramm.i18n.I18n;

import static de.bernhardjung.dsaprogramm.daten.xml.NamenlisteXmlTags.*;

/**
 * Ein Wort, das zum Bilden eines Heldennamens verwendet wird, z.B. ein Vorname
 * wie Alrik oder ein Nachname. Es kann aber auch ein Bindewort wie ibn oder
 * von sein.
 */
public class Name implements Comparable< Name > {

    /**
     * Das Suffix, dass an den Lokalisierungschschluessel (ID) angehaengt wird,
     * um den lokalisierten Namen zu ermitteln.
     */
    private final String SUFFIX_SCHLUESSEL_NAME = ".Name";
    
    /**
     * Eine Id, ueber die der Name lokalisiert werden kann. Diese Angabe ist
     * optional, man kann einen Namen auch direkt eingeben, ohne dass er
     * lokalisiert werden kann, in dem Fall ist der Schluessel {@code null}.
     */
    private String lokalisierungsschluessel;

    /** Das eigentlichen Wort, das fuer den Namen verwendet wird. */
    private String name;

    public Name() {
        
        lokalisierungsschluessel = null;

        name = "";
    }

    /**
     * Sortiert Namen nach Name und Lokalisierungsschluessel.
     */
    public int compareTo( final Name other ) {
        
        int ergebnis = Util.compare( getName(), other.getName() );
        
        if ( ergebnis != 0 ) {
            
            return ergebnis;
        }
        
        String ls1 = getLokalisierungsschluessel();
        
        String ls2 = other.getLokalisierungsschluessel();
        
        if ( ( null == ls1 ) &&  ( null == ls2 ) ) {
            
            return 0;
        
        } else if ( null == ls1 ) {
            
            return 1;
            
        } else if ( null == ls2 ) {
            
            return -1;
            
        } else {
            
            return ls1.compareTo( ls2 );
            
        }
    }

    
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime
                * result
                + ( ( lokalisierungsschluessel == null ) ? 0
                        : lokalisierungsschluessel.hashCode() );
        result = prime * result + ( ( name == null ) ? 0 : name.hashCode() );
        return result;
    }

    @Override
    public boolean equals( Object obj ) {
        if ( this == obj )
            return true;
        if ( obj == null )
            return false;
        if ( getClass() != obj.getClass() )
            return false;
        final Name other = (Name) obj;
        if ( lokalisierungsschluessel == null ) {
            if ( other.lokalisierungsschluessel != null )
                return false;
        } else if ( !lokalisierungsschluessel
                .equals( other.lokalisierungsschluessel ) )
            return false;
        if ( name == null ) {
            if ( other.name != null )
                return false;
        } else if ( !name.equals( other.name ) )
            return false;
        return true;
    }

    /**
     * @return Den Lokalisierungsschluessel, wenn es einen gibt. Wenn es keinen
     *         Schluessel gibt wird {@code null} geliefert.
     */
    public String getLokalisierungsschluessel() {
        
        return lokalisierungsschluessel;
    }
    
    /**
     * @return Den Namen. Es wird nie {@code null} geliefert.
     */
    public String getName() {
        
        return name;
    }
    
    /**
     * Legt den Loaklisierungsschluessel des Elements fest. Beim setzen des
     * schluessen wird der zugehoerige Name direkt aus der Lokalisierungsdatei
     * geladen. Wird kein Text in der Lokalisierungsdatei gefunden, dann wird
     * der schluessel als name verwendet.
     * 
     * @param schluessel
     *            Ein Schluessel aus der Lokalisierungsdatei.
     */
    public void setLokalisierungsschluessel( String schluessel ) {
        
        if ( null != schluessel ) {

            lokalisierungsschluessel = schluessel.trim();

            schluessel = lokalisierungsschluessel + SUFFIX_SCHLUESSEL_NAME;
            
            if ( I18n.hatLokalisierungstext( schluessel ) ) {
                
                name = I18n.i18n( schluessel );
            
            } else {
                
                name = schluessel;
            }
        
        } else {
            
            lokalisierungsschluessel = null;
        }
    }

    /**
     * Legt den Namen fest, wenn dieser nicht ueber einen Schluessel aus der
     * Lokalisierungdatei geladen wird.
     * 
     * @param derName
     *            Der eigentliche Name. Ist das Argument {@code null} wird es
     *            einfach ignoriert.
     */
    public void setName( final String derName ) {
        
        if ( null != derName ) {
            
            name = derName;
        }
    }

    /**
     * \return Den Namen.
     */
    @Override
    public String toString() {
        
        return name;
    }
    
    /**
     * @return Die XML representation des Elements.
     */
    public IXMLElement toXml() {
        
        final IXMLElement element = new XMLElement( TAG_NAME );
        
        if ( null != lokalisierungsschluessel ) {
            
            element.setAttribute( 
                    ATTR_LOKALISIERUNGSSCHLUESSEL, 
                    lokalisierungsschluessel );
        }
        
        if ( ( null == lokalisierungsschluessel )
                || ( ! I18n.hatLokalisierungstext( lokalisierungsschluessel ) )
                || ( ! I18n.i18n( lokalisierungsschluessel ).equals( name ) ) ){           
            
            element.setAttribute( ATTR_NAME, name );
        }
        
        return element;
    }
}
