/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.basis.namen;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

import static de.bernhardjung.dsaprogramm.daten.xml.NamenlisteXmlTags.*;
import de.bernhardjung.dsaprogramm.gui.Util;
import de.bernhardjung.dsaprogramm.i18n.I18n;

/**
 * Eine List von Namen einer bestimmten Kategorie z.B. nur weibliche Vornamen
 * oder Bindewoerter fuer adlige Namen.
 */
public class Namenliste implements Comparable< Namenliste > {
        
    private static final String SUFFIX_SCHLUESSEL_BEZEICHNUNG = ".Bezeichnung";
    
    /**
     * Die Bezeichnung der Liste z.B. "Vornamen (weiblich)".
     */
    private String bezeichnung;

    /** 
     * Eine Id zum identifizieren der Liste. Die Id muss nur unter den Listen
     * die zu einer Factory gehoeren eindeutig sein.
     */
    private String id;
    
    /** Die eigentliche Liste der Namen. */
    private List< Name > namen;

    /** Ein Zufallselement, um einen Namen aus der Liste zu waehlen. */
    private Random random;
    
    public Namenliste( final String dieId ) {
        
        bezeichnung = getLokalisierteBezeichnung( dieId );

        id = dieId;
        
        namen = new ArrayList< Name >();
        
        random = new Random();
    }
    
    /**
     * Sortiert die Liste nach Bezeichnung und Id.
     */
    public int compareTo( Namenliste o ) {
        
        int ergebnisId = getId().compareTo( o.getId() );
        
        if ( 0 == ergebnisId )
        {
            return 0;
        }
        
        int ergebnisBezeichnung = 
            Util.compare( getBezeichnung(),o.getBezeichnung() );
        
        if ( 0 != ergebnisBezeichnung )
        {
            return ergebnisBezeichnung;
        }
        
        return ergebnisId;
    }
    
    @Override
    public boolean equals( Object obj ) {
        
        if ( null == obj )
        {
            return false;
        }
        
        if ( this == obj )
        {
            return true;
        }
        
        if ( getClass() != obj.getClass() )
        {
            return false;
        }
        
        Namenliste andere = (Namenliste) obj;
        
        return 0 == compareTo( andere );
    }
    
    /**
     * @return Die Bezeichnung der Liste.
     */
    public String getBezeichnung() {
        
        return bezeichnung;
    }
    
    /** 
     * @return Die ID der Liste. Die ID ist nur unter den Listen einer Factory
     *         eindeutig.
     */
    public String getId()
    {
        return id;
    }
    
    /**
     * @return Liefert die lokalisierte Bezeichnung der Liste oder die Id, wenn
     *         es keine Lokalisierung gibt.
     */
    private String getLokalisierteBezeichnung( final String id ) {
        
        String schluesselBezeichnung = id + SUFFIX_SCHLUESSEL_BEZEICHNUNG;
        
        if ( I18n.hatLokalisierungstext( schluesselBezeichnung ) ) {
            
            return I18n.i18n( schluesselBezeichnung );
        
        } else {
            
            return id;
        }
    }
    
    /**
     * @return Die Namen in der Liste.
     */
    public List< Name > getNamen() {
        
        return new ArrayList< Name >( namen );
    }
    
    /**
     * @return Liefert einen Namen aus der Liste. Liefert einen leeren String,
     *         wenn die Liste leer ist.
     */
    public String getZufaelligenNamen() {
        
        if ( namen.isEmpty() ) {
            
            return "";
        }
        
        int size = namen.size();
        
        int index = random.nextInt( size );
        
        return namen.get( index ).toString();
    }
    
    /**
     * @return Den Hash Code der ID.
     */
    @Override
    public int hashCode() {
        
        return id.hashCode();
    }
    
    /**
     * Legt die Bezeichnung der Liste fest.
     * 
     * @param dieBezeichnung Wird {@code null} oder der leere String uebergeben
     *          wird das Argument ignoriert und versucht einen Text aus der
     *          Lokalisierungsdatei zu laden.
     */
    public void setBezeichnung( final String dieBezeichnung ) {
        
        if ( ( null != dieBezeichnung ) 
                && ( 0 < dieBezeichnung.trim().length() ) ) {
            
            bezeichnung = dieBezeichnung.trim();
        
        } else {
            
            bezeichnung = getLokalisierteBezeichnung( id );
        }
    }
    
    /**
     * Legt die Namen der Liste fest.
     * 
     * @param neueNamen Die Namen der Liste.
     */
    public void setNamen( final List< Name > neueNamen ) {
        
        namen.clear();
        
        namen.addAll( neueNamen );
    }
    
    /**
     * @return Die Bezeichnung der Liste.
     */
    @Override
    public String toString() {
        
        return bezeichnung;
    }
    
    /**
     * Wandelt die Liste in eine XML Repräsentation um, damit sie so gespeichert
     * werden kann.
     * 
     * @return Die XML Repräsentation der Liste.
     */
    public IXMLElement toXml() {
    
        IXMLElement element = new XMLElement( TAG_NAMENLISTE );
        
        element.setAttribute( ATTR_ID, id );
        
        final String schluesselBezeichnung = id + SUFFIX_SCHLUESSEL_BEZEICHNUNG;
            
        if ( ( ! id.equals( bezeichnung ) )
                && ( ! I18n.hatLokalisierungstext( schluesselBezeichnung ) ) 
                    || ( ! I18n.i18n( schluesselBezeichnung ).equals( bezeichnung ) ) ) {
            
            element.setAttribute( ATTR_BEZEICHNUNG, bezeichnung );
        }
        
        for ( final Name name : namen ) {
            
            element.addChild( name.toXml() );
        }
        
        return element;
    }    
}
