/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.daten.factorys.bausteine;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_BESCHREIBUNG;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_VARIANTE;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_VARIANTE_AENDERBAR;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_GELAENDEKUNDE;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_SONDERFERTIGKEIT_BONI;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_WERTE;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import net.n3.nanoxml.IXMLElement;
import de.bernhardjung.dsaprogramm.basis.Kulturkunde;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitBonusStandard;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitWertBonus;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;
import de.bernhardjung.dsaprogramm.zauber.Repraesentation;

/**
 * Erzeugt die Sonderfertigkeiten-Boni aus dem XML-Dokument.
 *
 * @author Bernhard Jung
 *
 */
public class FactorySonderfertigkeitBoniIn extends FactorySonderfertigkeitIn {

    /**
     * Ermittelt die Variante aus einem Wert-Element des XML-Dokuments.
     *
     * @param wertElement
     *            das auf eine Variante hin untersucht werden soll.
     * @return eine Variante oder {@code null} wenn das Element keine Variante
     *         vorgibt.
     */
    private final Object ermitteleVariante( final IXMLElement wertElement ) {

        final String varianteId =
            wertElement.getAttribute( ATTR_VARIANTE );

        if ( null == varianteId ) {
            return null;
        }

        if ( varianteId.startsWith( "Kulturkunde." ) ) {

            return Kulturkunde.getKulturkunde( varianteId );

        } else if ( varianteId.startsWith( "Repraesentation." ) ) {

            return Repraesentation.getRepraesentation( varianteId );

        } else if ( I18n.hatLokalisierungstext( varianteId ) ) {

            return I18n.i18n( varianteId );

        } else  {

            return varianteId;

        }
    }

    /**
     * Erzeugt aus einem Element des XML-Dokuments einen konkreten
     * Sonderfertigkeit-Bonus.
     *
     * @param bonus
     *            Das XML-Element das den Bonus repraesentiert.
     * @return den aus dem Element ermittelten Bonus.
     * @throws IllegalArgumentException
     *             wenn der {@code bonus} unbekannt ist und kein konkreter
     *             Bonus ermittelt werden kann.
     */
    private final SonderfertigkeitBonus erzeugeBonus( final IXMLElement bonus ) {

        final String name = bonus.getName();

        if ( name.equals( TAG_WERTE ) ) {

            // Es gibt also eine Liste von Werten aus denen der Benutzer
            // waehlen darf.
            final Vector< IXMLElement > wertElemente = bonus.getChildren();

            // Hier packen wir die Werte rein.
            final List< SonderfertigkeitWertBonus > werte =
                new ArrayList< SonderfertigkeitWertBonus >( wertElemente.size() );

            // Fuer den Fall dass die Werte Varianten haben muss noch
            // festgestellt werden ob die Variante vom Benutzer geaendert
            // werden darf
            final boolean varianteAenderbar;

            final String varianteAenderbarString =
                bonus.getAttribute( ATTR_VARIANTE_AENDERBAR );

            if ( null == varianteAenderbarString ) {

                varianteAenderbar = false;

            } else {

                varianteAenderbar = Boolean.valueOf( varianteAenderbarString );

            }

            // Jetzt die einzelnen Werte erzeugen
            for ( final IXMLElement wertElement : wertElemente ) {

                final Sonderfertigkeit sonderfertigkeit =
                    ermitteleSonderfertigkeit( wertElement );

                // Dann die Variante ermitteln.
                final Object variante = ermitteleVariante( wertElement );

                // zuletzt den Wert erzeugen.
                final SonderfertigkeitWert wert;

                if ( null != variante ) {

                    wert = new SonderfertigkeitWert(
                            sonderfertigkeit,
                            variante,
                            varianteAenderbar );
                } else {

                    wert = new SonderfertigkeitWert( sonderfertigkeit );
                }

                werte.add( new SonderfertigkeitWertBonus(
                        wert,
                        wertElement.getAttribute( ATTR_VARIANTE ) ) );
            }

            // Jetzt noch eine Beschreibung ermitteln, so eine da ist.
            final String beschreibung = bonus.getAttribute( ATTR_BESCHREIBUNG );

            return new SonderfertigkeitBonusStandard(
                    beschreibung,
                    varianteAenderbar,
                    werte );/* );*/

        }/* else if ( name.equals( TAG_VARIANTEN ) ) {

            // Der Benutzer darf also aus allen Varianten einer
            // Sonderfertigkeit waehlen.

            // Erst die Sonderfertigkeit ermitteln
            final String sonderfertigkeitId =
                bonus.getAttributeValue( ATTR_SONDERFERTIGKEIT );

            final Sonderfertigkeit sonderfertigkeit =
                Sonderfertigkeit.getSonderfertigkeit( sonderfertigkeitId );

            // Aus diesen Varianten darf gewaehlt werden.
            final List< Object > varianten = sonderfertigkeit.getVarianten();

            // Jetzt die Werte erzeugen aus denen zu waehlen ist.
            final List< SonderfertigkeitWert > werte =
                new ArrayList< SonderfertigkeitWert >( 1 );

            werte.add( new SonderfertigkeitWert( sonderfertigkeit,
                                                 varianten.get( 0 ) ) );

            // Jetzt noch eine evtl Beschreibung ermitteln.
            final String beschreibung = ermittleBeschreibung( bonus );

            // Fertig.
            return new SonderfertigkeitBonusStandard( beschreibung, werte );

        }*/ else if ( name.equals( TAG_GELAENDEKUNDE ) ) {

            return SonderfertigkeitBonus.GELAENDEKUNDE;

        } else {

            throw new IllegalArgumentException( String.format(
                    "Der Sonderfertigkeit-Bonus Typ %1$s ist unbekannt.",
                    name ) );
        }
    }

    /**
     * Erzeugt aus dem XML-Dokument die Boni auf die Sonderfertigkeiten.
     *
     * @param wurzelElement
     *            des XML-Dokuments.
     * @return die aus dem XML-Dokument ermittelten Boni auf die
     *         Sonderfertigkeiten.
     */
    public final List< SonderfertigkeitBonus > erzeugen(
            final IXMLElement wurzelElement ) {

        final List< SonderfertigkeitBonus > boni =
            new ArrayList< SonderfertigkeitBonus >();

        final IXMLElement boniElement =
            wurzelElement.getFirstChildNamed( TAG_SONDERFERTIGKEIT_BONI );

        if ( null != boniElement ) {

            for ( final IXMLElement bonus : boniElement.getChildren() ) {

                boni.add( erzeugeBonus( bonus ) );
            }
        }

        return boni;
    }
}
