package de.bernhardjung.dsaprogramm.gui.meta;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractListModel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.bernhardjung.dsaprogramm.i18n.I18n;

/**
 * Ein Editor um eine Liste von Formatvorlagen fuer den Namen eines Helden zu
 * waehlen.
 * Der Editor besitzt links eine Liste in der die gewaehlten Formate angezeigt
 * werden und rechts ein Textfeld in das man ein neues Format eingeben kann.
 */
class PanelMetaNameFormate extends JPanel {

    /**
     * Der Text auf dem Knopf mit dem man ein Format aus der Liste der Formate
     * entfernen kann.
     */
    private static final String ENTFERNEN = 
        I18n.i18n( "PanelMetaNameFormate.Entfernen" );
    
    /**
     * Der Text bei dem Textfeld in das man ein neues Format eingeben kann.
     */
    private static final String FORMAT =
        I18n.i18n( "PanelMetaNameFormate.Format" );
    
    /**
     * Der Text auf dem Knopf mit dem man das im Textfeld eingegebene Format in
     * die Liste der Formate uebernimmt.
     */
    private static final String HINZUFUEGEN =
        I18n.i18n( "PanelMetaNameFormate.Hinzufuegen" );
    
    private JTextField format;
    
    private FormateListModel formate;
    
    public PanelMetaNameFormate() {
    
        format = new JTextField();
        format.addActionListener( new ActionListener() {
            public void actionPerformed( ActionEvent e ) {
                hinzufuegenFormat();
            }
        } );
        
        formate = new FormateListModel();
        
        aufbauenBenutzeroberflaeche();
    }
    
    /**
     * Entfernt das Format von der gegebenen position in der Liste.
     * 
     * @param index des zu entfernenden Formats.
     */
    private void entferneFormatAusListe( final int index ) {
        
        if ( -1 == index ) {
            
            return;
        }
        
        formate.entfernen( index );
    }
    
    private void aufbauenBenutzeroberflaeche() {
        
        final String zeilen = "F:P:G";
        
        final String spalten = "F:100px:G(0.2), 4dlu, F:P:G(0.8)";
        
        final FormLayout layout = new FormLayout( spalten, zeilen );
        
        final DefaultFormBuilder builder = 
            new DefaultFormBuilder( layout, this );
        builder.setDefaultDialogBorder();
        
        builder.append( erzeugeLinkeSpalte() );
        
        builder.append( erzeugeRechteSpalte() );
    }
    
    /**
     * @return Den linken Teil des Editors mit der Liste und dem Knopf darunter.
     */
    private JComponent erzeugeLinkeSpalte() {
        
        final JList formateListe = new JList( formate );
        
        final JScrollPane scrollPane = new JScrollPane( formateListe );

        final JButton entfernen = new JButton( ENTFERNEN );
        entfernen.addActionListener( new ActionListener() {
            public void actionPerformed( ActionEvent e ) {
                entferneFormatAusListe( formateListe.getSelectedIndex() );
            }
        } );
        
        final JPanel panel = new JPanel( new BorderLayout() );
        
        panel.add( scrollPane, BorderLayout.CENTER );
        
        panel.add( entfernen, BorderLayout.PAGE_END );
        
        return panel;
    }
    
    /**
     * @return Der rechte Teil des Editors mit dem Texfeld und den Knopf 
     *         darunter.
     */
    private JComponent erzeugeRechteSpalte() {
        
        final JButton hinzufuegen = new JButton( HINZUFUEGEN );
        hinzufuegen.addActionListener( new ActionListener() {
            public void actionPerformed( ActionEvent e ) {
                hinzufuegenFormat();
            }
        } );
        
        final FormLayout layout = new FormLayout( "P, 4dlu, F:P:G" );
        
        final DefaultFormBuilder builder = new DefaultFormBuilder( layout );
        
        builder.append( FORMAT, format );
        
        builder.append( hinzufuegen, 3 );
        
        return builder.getPanel();
    }
    
    /**
     * @return Die vom Benutzer gewaehlten Formate.
     */
    public List< String > getFormate() {
        
        return formate.getFormate();
    }

    /**
     * Fuegt das aktuell im Textfeld eingegebenen Format in die Liste der 
     * Formate ein.
     */
    private void hinzufuegenFormat() {
        
        final String format = this.format.getText().trim();
        
        if ( format.equals( "" ) ) {
            
            return;
        }
        
        formate.hinzufuegen( format );
    }
    
    /** Loescht alle Eingaben aus dem Editor. */
    public void reset()
    {
        format.setText( "" );
        
        formate.reset();
    }
    
    /**
     * @param neueFormate die im Editor angezeigt werden sollen.
     */
    public void setFormate( final List< String > neueFormate ) {
        
        formate.setFormate( neueFormate );
    }
    
    private class FormateListModel extends AbstractListModel {
        
        private final List< String > formate = new ArrayList< String >();

        /**
         * Entfernt das Format von der gegebenen Position in der Liste.
         * 
         * @param index des zu entferneneden Formats.
         */
        public void entfernen( final int index ) {
            
            formate.remove( index );
            
            fireIntervalRemoved( this, index, index );
        }
        
        public Object getElementAt( int index ) {
            return getFormat( index );
        }
        
        /**
         * Liefert das Formate von der gegebenen Position im Model.
         * 
         * @param index des gewuenschten Formats.
         * @return
         */
        public String getFormat( int index ) {
            
            return formate.get( index );
        }

        /**
         * @return Ein Liste aller Formate im Model.
         */
        public List< String > getFormate() {
            
            return new ArrayList< String >( formate );
        }
        
        public int getSize() {
            return formate.size();
        }
        
        /**
         * Fuegt dem Model ein neues Format hinzu, aber nur wenn das Format noch 
         * nicht vorhanden ist.
         * 
         * @param format Das neue Format.
         */
        public void hinzufuegen( final String format ) {
            
            if ( formate.contains( format ) ) {
                
                return;
            }
            
            int index = formate.size();
            
            formate.add( format );
            
            fireIntervalAdded( this, index, index );
        }
        
        /**
         * Leert den Inhalt des Models.
         */
        public void reset()
        {
            int index = Math.max( 0, -1 + formate.size() );
            
            formate.clear();
            
            fireIntervalRemoved( this, 0, index );
        }
        
        /**
         * Legt die Formate im Model fest.
         * 
         * @param neueFormate die neuen Formate.
         */
        public void setFormate( final List< String > neueFormate ) {
            
            formate.clear();
            
            formate.addAll( neueFormate );
            
            fireContentsChanged( this, 0, -1 + formate.size() );
        }
    }
}
