/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.sonderfertigkeiten;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.util.FehlendesArgumentException;

/**
 * Repraesentiert eine Sonderfertigkeit die ein Held besitzt.
 * Hier werden alle fuer den Held relevanten Informationen zur
 * Sonderfertigkeit gespeichert, d.h die Variante.
 *
 * @author Bernhard Jung
 */
public class SonderfertigkeitWert implements Cloneable {

    /** Der Held dem der Wert gehoert. Kann auch {@code null} sein. */
    private Held held;

    /** Die AP die der Held bezahlt hat um die Sonderfertigkeit zu erlernen. */
    private int lernkosten;

    /**
     * Der Name des Wertes. Setzt sich aus Name der Sonderfertigkeit und
     * Varianten zusammen.
     */
    private String name;

    /** Die Sonderfertigkeit des Wertes. */
    private final Sonderfertigkeit sonderfertigkeit;

    /**
     * Die Variante der Sonderfertigkeit, falls die Sonderfertigkeit
     * Varianten besitzt.
     */
    private Object variante;

    /**
     * Gibt an ob die Variante des Wertes geaendert werden darf. Bei
     * Sonderfertigkeiten die ein Held durch Kultur oder Profession erhalten
     * hat ist das normalerweise nicht der Fall.
     */
    private boolean varianteAenderbar;

    /**
     * Erzeugt einen neuen Wert.
     *
     * @param dieSonderfertigkeit auf die sich der Wert bezieht.
     * @throws IllegalArgumentException wenn die Sonderfertigkeit Varianten
     *      besitzt. In diesem Fall muss der Konstruktor {@link
     *      #SonderfertigkeitWert(Sonderfertigkeit, Object, boolean)} verwendet
     *      werden.
     */
    public SonderfertigkeitWert( final Sonderfertigkeit dieSonderfertigkeit ) {

        if ( dieSonderfertigkeit.hatVarianten() ) {

            throw new FehlendesArgumentException( String.format(
                    "Die Sonderfertigkeit %1$s besitzt Varianten. Bei der " +
                    "Erzeugung eines Wertes muss ein Variante angegeben " +
                    "werden. Bitte benutzen Sie den Konstruktor " +
                    "SonderfertigkeitWert( Sonderfertigkeit, Object ).",
                    dieSonderfertigkeit.getName() ) );
        }

        lernkosten = 0;

        name = dieSonderfertigkeit.getName();

        sonderfertigkeit = dieSonderfertigkeit;

        variante = "";

        varianteAenderbar = false;
    }

    /**
     * Erzeugt einen neuen Wert dessen Variante nachtraeglich geaendert
     * werden kann.
     *
     * @param dieSonderfertigkeit auf die sich der Wert bezieht.
     * @param variante der Sonderfertigkeit. Das Argument darf nicht
     *      {@code null} sein, siehe {@link #setVariante(Object)}.
     */
    public SonderfertigkeitWert(
            final Sonderfertigkeit dieSonderfertigkeit,
            final Object variante ) {

        this( dieSonderfertigkeit, variante, true );
    }

    /**
     * Erzeugt einen neuen Wert.
     *
     * @param dieSonderfertigkeit auf die sich der Wert bezieht.
     * @param variante der Sonderfertigkeit. Das Argument darf nicht
     *      {@code null} sein, siehe {@link #setVariante(Object)}.
     * @param istVarianteAenderbar gibt an ob die Variante nachtraeglich
     *      geaendert werden darf.
     */
    public SonderfertigkeitWert(
            final Sonderfertigkeit dieSonderfertigkeit,
            final Object variante,
            final boolean istVarianteAenderbar ) {

        if ( ! dieSonderfertigkeit.hatVarianten() ) {

            throw new IllegalArgumentException( String.format(
                    "Die Sonderfertigkeit %1$s besitzt keine Varianten. Bei " +
                    "der Erzeugung des Wertes wurde aber eine Variante " +
                    "angegeben. Bitte benutzen Sie den Konstruktor " +
                    "SonderfertigkeitWert( Sonderfertigkeit ).",
                    dieSonderfertigkeit.getName() ) );
        }

        lernkosten = 0;

        // Name wird von setVariante(String) gesetzt, siehe unten.
        // name = dieSonderfertigkeit.getName();

        sonderfertigkeit = dieSonderfertigkeit;

        // erst auf true setzen, damit setVariante aufgerufen werden kann.
        varianteAenderbar = true;
        setVariante( variante );

        varianteAenderbar = istVarianteAenderbar;
    }

    /**
     * @return eine shallow Kopie des Wertes.
     */
    @Override
    public SonderfertigkeitWert clone() {
        try {
            return (SonderfertigkeitWert) super.clone();
        } catch (final CloneNotSupportedException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Die Methode wurde ueberschrieben, so dass Werte mit gleichem Namen als
     * gleich gelten.
     *
     * @param obj das andere Objekt mit dem verglichen wird.
     */
    @Override
    public boolean equals( final Object obj ) {

        if ( this == obj ) {
            return true;
        }

        if ( obj == null ) {
            return false;
        }

        final SonderfertigkeitWert other = (SonderfertigkeitWert) obj;

        return name.equals( other.name );
    }

    /**
     * @return die AP-Kosten fuer die Sonderfertigkeit.
     */
    public double getApKosten() {
        return sonderfertigkeit.getApKosten( this );
    }

    /**
     * @param held der die Sonderfertigkeit bezahlen muss.
     * @return die AP die der Held fuer die Sonderfertigkeit bezahlen muss.
     */
    public final int getApKosten( final Held held ) {
        return sonderfertigkeit.getApKosten( held, this );
    }

    /**
     * @param held
     *            der die Sonderfertigkeit erlernen will.
     * @param lernmethode
     *            die zum Erlernen der Sonderfertigkeit verwendet werden soll.
     * @return die AP-Kosten die der Held zum lernen der Sonderfertigkeit zahlen
     *         muss.
     */
    public int getApKosten( final Held held, final Lernmethode lernmethode ) {
        return sonderfertigkeit.getApKosten( held, this, lernmethode );
    }

    /**
     * @see #getGpKosten(Held)
     * @return die Standard GP-Kosten fuer den Wert. Kann der Wert ohne den Held
     *         nicht genau bestimmt werden wird irgendein Wert geliefert.
     */
    public final int getGpKosten() {
        return sonderfertigkeit.getGpKosten( this );
    }

    /**
     * @param held der die Sonderfertigkeit bezahlen muss.
     * @return die GP die der Held fuer die Sonderfertigkeit bezahlen muss.
     */
    public final int getGpKosten( final Held held ) {
        return sonderfertigkeit.getGpKosten( held, this );
    }

    /**
     * @return die AP die ein Held aufgebracht hat um die Sonderfertigkeit zu
     *         erlernen.
     */
    public final int getLernkosten() {
        return lernkosten;
    }

    /**
     * @return die Bezeichnung des Wertes, z.B.
     *      'Talentspezialisierung Klettern (Freiklettern)'
     */
    public final String getName() {
        return name;
    }

    /**
     * @return die Sonderfertigkeit des Wertes.
     */
    public final Sonderfertigkeit getSonderfertigkeit() {
        return sonderfertigkeit;
    }

    /**
     * @return die gewaehlte Variante der Sonderfertigkeit oder {@code null}
     *      falls die Sonderfertigkeit keine Varianten besitzt.
     */
    public final Object getVariante() {
        return variante;
    }

    @Override
    public int hashCode() {
        return name.hashCode();
    }

    /**
     * @return {@code true} wenn die Variante des Wertes geaendert werden darf,
     *      sonst {@code false}.
     */
    public final boolean istVarianteAenderbar() {
        return varianteAenderbar;
    }

    /**
     * Prueft ob ein Held alle Voraussetzungen erfuellt um die Sonderfertigkeit
     * einsetzen zu koennen.
     *
     * @param held der die Sonderfertigkeit einsetzen will.
     * @return eine Beschreibung der bestehenden Probleme wenn der Held nicht
     *      alle Voraussetzungen erfuellt.
     */
    public final KonfliktBeschreibung pruefen( final Held held ) {
        return sonderfertigkeit.pruefen( held, this );
    }

    /**
     * Bestimmt den Helden der den Wert besitzt. Der Held wird im folgenden
     * ueber Aenderungen an diesem Wert informiert. Dazu wird die Methode
     * {@link Held#notifyObservers()} aufgerufen.
     *
     * @param derHeld der den Wert bestizt.
     */
    public final void setHeld( final Held derHeld ) {
        held = derHeld;
    }

    /**
     * @param ap
     *            die ein Held aufgebracht hat, um die Sonderfertigkeit zu
     *            erlernen.
     */
    public final void setLernkosten( final int ap ) {
        lernkosten = ap;
    }

    /**
     * Bestimmt die Variante der Sonderfertigkeit.
     *
     * @see #istVarianteAenderbar()
     * @param dieVariante der Sonderfertigkeit. Das Objekt muss mittels
     *      {@link Object#toString()} sinnvoll in eine String umgewandelt
     *      werden koennen.
     * @throws IllegalArgumentException wenn {@code null} uebergeben wird.
     * @throws UnsupportedOperationException wenn die Variante nicht aenderbar
     *      ist oder die Sonderfertigkeit keine Varianten besitzt.
     */
    public final void setVariante( final Object dieVariante ) {

        if ( null == dieVariante ) {

            throw new IllegalArgumentException( String.format(
                    "Als Variante fuer die Sonderfertigkeit %1$s wurde null " +
                    "uebergeben. Das ist nicht erlaubt.",
                    getName() ) );
        }

        if ( ! sonderfertigkeit.hatVarianten() ) {

            throw new UnsupportedOperationException( String.format(
                    "Die Sonderfertigkeit %1$s besitzt keine Varianten.",
                    sonderfertigkeit.getName() ) );
        }

        if ( ! varianteAenderbar ) {

            throw new UnsupportedOperationException( String.format(
                    "Die Variante von %1$s darf nicht geaendert werden. ",
                    getName() ) );
        }

        variante = dieVariante;

        name = sonderfertigkeit + " (" + variante + ")";

        if ( null != held ) {
            held.notifyObservers( sonderfertigkeit );
        }
    }

    /**
     * @param aenderbar
     *            {@code true}, wenn die Variante aenderbar sein soll, sonst
     *            {@code false}.
     */
    public void setVarianteAenderbar( final boolean aenderbar ) {
        varianteAenderbar = aenderbar;
    }

    /**
     * @return {@link #getName()}
     */
    @Override
    public String toString() {
        return getName();
    }
}
