/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.talente;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

/**
 * Repraesentiert eine Sprache.
 * Sprachen besitzten gegenueber herkoemmlichen Talenten noch eine
 * Komplexitaet, welche die Hoechststufe angibt bis zu der die Sprache
 * gesteigert werden kann, und eine Liste von Schriften die der Sprache
 * zugeordnet sind. Weiter ist jede Sprache einer Sprachfamilie zugeordnet.
 * Zu jeder existierenden Sprache gibt es in der Klasse {@link Talent} eine
 * statische Konstante.
 *
 * @author Bernhard Jung
 */
public class Sprache extends Talent {

    /**
     * Die Komplexitaet bis zu der die Sprache gesteigert werden kann.
     */
    private int komplexitaet;

    /** Die Schriften in der die Sprache geschrieben wird. */
    private final Collection< Schrift > schriften;

    /** Die Sprachfamilie der Sprache. */
    private Sprachfamilie sprachfamilie;

    /**
     * @param dieId der Sprache.
     */
    Sprache( final String dieId ) {

        super( dieId, Talentgruppe.SPRACHEN, Typ.SPEZIAL );

        komplexitaet = -1;

        schriften = new HashSet< Schrift >( 1 );
    }

    /**
     * Fuegt der Sprache eine Schrift, in der die Sprache niedergeschrieben
     * wird, hinzu.
     *
     * @param schrift in der die Sprache geschrieben wird.
     * @throws AssertionError wenn eine Schrift zum zweiten mal eingefuegt wird
     *      oder das Argument {@code null} ist.
     */
    void einfuegen( final Schrift schrift ) {

        assert( null != schrift ) :
            "Das Argument fuer schrift darf nicht null sein.";

        assert( ! schriften.contains( schrift ) ) : String.format(
                "Schrift %1$s sollte zum zweiten mal zu Sprache %2$s " +
                "hinzugefuegt werden.",
                schrift.getName(),
                getName() );

        schriften.add( schrift );
    }

    /**
     * @return die Komplexitaet bis zu der die Sprache gesteigert werden kann.
     */
    public final int getKomplexitaet() {
        return komplexitaet;
    }

    /**
     * @return eine Liste der Schriften in der die Sprache geschrieben wird.
     */
    public final List< Schrift > getSchriften() {
        return new ArrayList< Schrift >( schriften );
    }

    /**
     * @return die Sprachfamilie der Sprache.
     */
    public final Sprachfamilie getSprachfamilie() {
        return sprachfamilie;
    }

    /**
     * Bestimmt die Komplexitaet der Sprache.
     * Die Methode sollte nur einmal aufgerufen werden um die Sprache zu
     * initialisieren.
     *
     * @param dieKomplexitaet der Sprache.
     * @throws AssertionError wenn die Komplexitaet schon einmal gesetzt wurde.
     */
    void setKomplexitaet( final int dieKomplexitaet ) {

        assert( -1 == komplexitaet ) : String.format(
                "Die Komplexitaet von %1$s sollte ein zweites mal gesetzt " +
                "werden.\n" +
                "bestehnden Komplexitaet: %2$s\n" +
                "neue Komplexitaet: %3$s",
                getName(),
                komplexitaet,
                dieKomplexitaet );

        komplexitaet = dieKomplexitaet;
    }

    /**
     * Bestimmt die Sprachfamilie der Sprache.
     * Diese Methode sollte nur einmal aufgerufen werden um die Sprache zu
     * initialisieren.
     *
     * @param dieSprachfamilie der Sprache.
     * @throws AssertionError wenn die Sprache schon einer Sprachfamilie zugeordnet
     *      ist oder das Argument {@code null} ist.
     */
    void setSprachfamilie( final Sprachfamilie dieSprachfamilie ) {

        assert ( null != dieSprachfamilie ) :
            "dieSprachfamilie darf nicht null sein.";

        assert ( null == sprachfamilie ) : String.format(
            "Die Sprachfamilie der Sprache %1$s sollte zum zweiten mal " +
            "gesetzt werden.\n" +
            "bisherige Sprachfamilie: %2$s\n" +
            "neue Sprachfamilie: %3$s",
            getName(),
            sprachfamilie,
            dieSprachfamilie );

        sprachfamilie = dieSprachfamilie;
    }
}
