/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.vornachteile;

import java.util.ArrayList;
import java.util.List;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.util.Beobachtbar;
import de.bernhardjung.dsaprogramm.util.FehlendesArgumentException;

/**
 * Repraesentiert den (moeglichen) Wert eines Helden.
 *
 * @author Bernhard Jung
 *
 * TODO Der Held sollte auch ueber das Beobachter-Pattern ueber Aenderungen am
 *      Wert informiert werden.
 */
public class VorNachteilWert
extends Beobachtbar
implements Cloneable {

    /**
     * Der Held dessen Wert dies ist, falls der Wert einem Helden zugeordnet
     * ist.
     */
    private Held held;

    /**
     * Die AP die bezahlt wurden um eine Schlechte Eigenschaft abzubauen. An
     * Position i befinden sich die Kosten die bezahlt wurden um den Wert auf
     * Stufe i zu senken.
     */
    private List< Integer > kosten;

    /** Der Name des Wertes. */
    private String name;

    /** Markiert eine Spezielle Erfahrung bei einer Schlechten Eigenschaft. */
    private boolean spezielleErfahrung;

    /** Die Stufe des Wertes, so der Vor-/Nachteil denn Stufen hat. */
    private int stufe;

    /**
     * Die Variante des Wertes, falls der Vor-/Nachteil in Varianten
     * vorkommt.
     */
    private Object variante;

    /** Gibt an ob die variante des Wertes geaendert werden darf. */
    private boolean varianteAenderbar;

    /** Der Vor-/Nachteil auf den sich der Wert bezieht. */
    private final VorNachteil vorNachteil;

    /**
     * Erzeugt einen neuen Wert ohne Variante.
     *
     * @param derVorNachteil auf den sich der Wert bezieht.
     * @throws FehlendesArgumentException wenn der Vor-/Nachteil Varianten
     *      besitzt.
     */
    public VorNachteilWert( final VorNachteil derVorNachteil ) {

        if ( derVorNachteil.hatVarianten() ) {

            throw new FehlendesArgumentException( String.format(
                    "Der Vor-/Nachteil %1$s besitzt Varianten. Bei der " +
                    "Erzeugung eines Wertes muss ein Variante angegeben " +
                    "werden. Bitte benutzen Sie den Konstruktor " +
                    "VorNachteilWert( VorNachteil, Object ).",
                    derVorNachteil.getName() ) );

        }

        name = derVorNachteil.getName();

        kosten = new ArrayList< Integer >();

        spezielleErfahrung = false;

        stufe = derVorNachteil.getMindeststufe();

        variante = "";

        varianteAenderbar = false;

        vorNachteil = derVorNachteil;
    }

    /**
     * Erzeugt einen neuen Wert.
     *
     * @param derVorNachteil auf den sich der Wert bezieht.
     * @param variante des Vor-/Nachteils. Das Argument darf nicht
     *      {@code null} sein, siehe {@link #setVariante(Object)}.
     */
    public VorNachteilWert(
            final VorNachteil derVorNachteil,
            final Object variante ) {

        this( derVorNachteil, variante, true );
    }

    /**
     * Erzeugt einen neuen Wert.
     *
     * @param derVorNachteil
     *            auf den sich der Wert bezieht.
     * @param variante
     *            des Vor-/Nachteils. Das Argument darf nicht {@code null} sein,
     *            siehe {@link #setVariante(Object)}.
     * @param istVarianteAenderbar
     *            gib an ob die Variante des Wertes geaendert werden darf.
     */
    public VorNachteilWert(
            final VorNachteil derVorNachteil,
            final Object variante,
            final boolean istVarianteAenderbar ) {

        if ( ! derVorNachteil.hatVarianten() ) {

            throw new IllegalArgumentException( String.format(
                    "Der Vor-/Nachteil %1$s besitzt keine Varianten. Bei " +
                    "der Erzeugung des Wertes wurde aber eine Variante " +
                    "angegeben. Bitte benutzen Sie den Konstruktor " +
                    "VorNachteilWert( VorNachteil ).",
                    derVorNachteil.getName() ) );
        }

        // Name wird von setVariante(String) gesetzt, siehe unten.
        // name = derVorNachteil.getName();

        kosten = new ArrayList< Integer >();

        spezielleErfahrung = false;

        stufe = derVorNachteil.getMindeststufe();

        vorNachteil = derVorNachteil;

        varianteAenderbar = true;
        setVariante( variante );

        varianteAenderbar = istVarianteAenderbar;
    }

    /**
     * Erhoeht die Stufe des Wertes um einen Bonus.
     *
     * @param bonus um den die Stufe erhoeht werden soll.
     */
    public final void addiereStufe( final int bonus ) {

        if ( 0 != bonus ) {

            stufe += bonus;

            if ( null != held ) {
                held.notifyObservers( vorNachteil );
            }

            meldeBeobachtern( vorNachteil );
        }
    }

    /**
     * @return eine Kopie des Wertes.
     */
    @Override
    public final VorNachteilWert clone() {

        final VorNachteilWert kopie = (VorNachteilWert) super.clone();

        kopie.kosten = new ArrayList< Integer >( kosten );

        return kopie;
    }

    /**
     * Soll aufgerufen werden nachdem ein Held den Vor-/Nachteil erhalten hat.
     *
     * @param held
     *            der den Wert erhalten hat.
     */
    public final void eingefuegt( final Held held ) {
        vorNachteil.eingefuegt( held, this );
    }

    /**
     * Soll aufgerufen werden nachdem ein Held den Vor-/Nachteil verloren hat.
     *
     * @param held
     *            der den Wert verloren hat.
     */
    public final void entfernt( final Held held ) {
        vorNachteil.entfernt( held, this );
    }

    /**
     * Die Methode wurde ueberschrieben, so dass Werte mit gleicher
     * Vor-/Nachteil und gleicher Variante als gleich gelten.
     *
     * @param obj
     *            das andere Objekt mit dem verglichen wird.
     */
    @Override
    public final boolean equals( final Object obj ) {

        if ( this == obj ) {
            return true;
        }

        if ( null == obj ) {
            return false;
        }

        if ( getClass() != obj.getClass() ) {
            return false;
        }

        final VorNachteilWert other = (VorNachteilWert) obj;

        return vorNachteil.equals( other.vorNachteil )
            && variante.equals( other.variante );
    }

    /**
     * Liefert, bei einem Nachteil, die Kosten die zu zahlen sind, um den
     * Nachteil abzubauen.
     *
     * @param held der den Nachteil abbauen will.
     * @return die AP-Kosten, um den Nachteil abzubauen.
     */
    public final int getAbbaukosten( final Held held ) {
        return vorNachteil.getAbbaukosten( held, this  );
    }

    /**
     * Liefert die Standard-Kosten fuer den Wert. Wenn sich die Kosten ohne
     * den Helden nicht genau bestimmen lassen liefert diese Methode einen
     * beliebigen Wert.
     *
     * @see #getGpKosten(Held)
     * @return die GP-Kosten des Vor-/Nachteils.
     */
    public final int getGpKosten() {
        return vorNachteil.getGpKosten( this );
    }

    /**
     * Liefert die GP-Kosten die ein Held fuer den Vor-/Nachteil zahlen muss.
     *
     * @see #getGpKosten()
     * @param held der den Vor-/Nachteil bezahlen muss.
     * @return die GP-Kosten die der Held zahlen muss.
     */
    public final int getGpKosten( final Held held ) {
        return vorNachteil.getGpKosten( held, this );
    }

    /* *
     * @return die AP-Kosten die verwendet wurden um die Stufe der Schlechten
     *         Eigenschaft zu senken. Die Kosten um den Wert auf Stufe i zu
     *         senken stehen in der Liste an Position i.
     * /
    public final List< Integer > getKosten() {
        return new ArrayList< Integer >( kosten );
    }
*/
    /** @return den Namen des Wertes. */
    public final String getName() {
        return name;
    }

    /**
     * @param lernmethode
     *            mit der die Schlechte Eigenschaft gesenkt werden soll. Besitzt
     *            der Held eine Spezielle Erfahrung auf die Schlechte
     *            Eigenschaft so wird die uebergebene Lernmethode ignoriert.
     * @return bei einer Schlechten Eigenschaft, die AP, die zu zahlen sind, um
     *         den Wert um einen Punkt zu senken.
     */
    public final int getSenkungskosten( Lernmethode lernmethode ) {

        if ( spezielleErfahrung ) {

            lernmethode = Lernmethode.SPEZIELLE_ERFAHRUNG;
        }

        return vorNachteil.getSenkungskosten( lernmethode );
    }

    /**
     * @return die Stufe des Wertes. Wenn der Vor-/Nachteil keine Stufen hat
     *      wird ein beliebiger Wert zurueck gegeben.
     */
    public final int getStufe() {
        return stufe;
    }

    /**
     * @return die fuer den Wert gewaehlte Variante.
     * @throws UnsupportedOperationException
     *             wenn der Vor-/Nachteil keine Varianten besitzt.
     */
    public final Object getVariante() {

        if ( ! vorNachteil.hatVarianten() ) {

            throw new UnsupportedOperationException( String.format(
                    "Der Vor-/Nachteil %1$s bestizt keine Varianten",
                    vorNachteil.getName() ) );
        }

        return variante;
    }

    /** @return den Vor-/Nachteil auf den sich der Wert bezieht. */
    public final VorNachteil getVorNachteil() {
        return vorNachteil;
    }

    /**
     * @return {@code true}, wenn der Held eine Spezielle Erfahrung auf die
     *         Schlechte Eigenschaft erhalten hat, sonst {@code false}.
     */
    public final boolean hatSpezielleErfahrung() {
        return spezielleErfahrung;
    }

    /**
     * @return {@code true}, wenn die Variante des Wertes geaendert werden
     *         darf, sonst {@code false}. Wenn der Vor-/Nachteil keine Varianten
     *         besitzt, wird ein beliebiger Wert geliefert.
     */
    public final boolean istVarianteAenderbar() {
        return varianteAenderbar;
    }

    /**
     * Prueft ob der Held den Vor-/Nachteil besitzten darf oder ob es Konflikte
     * mit den Regeln gibt.
     *
     * @param held
     *            der den Wert waehlen will.
     * @return eine Beschreibung evtl. aufgetretener Regelkonflikte.
     */
    public final KonfliktBeschreibung pruefen( final Held held ) {
        return vorNachteil.pruefen( held, this );
    }

    /**
     * Senkt die Stufe des Wertes um einen Punkt.
     *
     * @param held
     *            der den Wert senken will.
     * @param lernmethode
     *            die zum Senken verwendet werden soll.
     * @return die AP die zu investieren sind, um die Stufe um einen Punkt zu
     *         senken.
     */
    private int senken(
            final Held held,
            final Lernmethode lernmethode ) {

        --stufe;

        int apKosten = 0;

        if ( held.hatVorgegeben( this ) ) {

            final int vorgegebeneStufe = held.getVorgegeben( this ).getStufe();

            if ( stufe < vorgegebeneStufe ) {

                apKosten = getSenkungskosten( lernmethode );

                setKosten( apKosten );

                spezielleErfahrung = false;
            }

        }

        return apKosten;
    }

    /**
     * Bestimmt den Helden dem der Wert gehoert. Bei Aenderungen an der
     * Variante oder der Stufe des Wertes wird {@link Held#notifyObservers()}
     * aufgerufen.
     *
     * @param derHeld dem der Wert gehoert.
     */
    public final void setHeld( final Held derHeld ) {
        held = derHeld;
    }

    /**
     * Legt die Kosten fest, die investiert wurden um den Wert auf die aktuelle
     * Stufe zu senken.
     *
     * @param apKosten
     *            die zum Senken aufgewendet wurden.
     */
    private void setKosten( final int apKosten ) {

        while ( kosten.size() <= stufe ) {

            kosten.add( 0 );
        }

        kosten.set( stufe, apKosten );
    }
/*
    /**
     * Legt die Kosten fest, die verwendet wurden um die Stufe der Schlechten
     * Eigenschaft zu senken.
     * @param neueKosten
     * /
    public final void setKosten( final List< Integer > neueKosten ) {
        kosten.clear();

        kosten.addAll( neueKosten );
    }
*/
    /**
     * Markiert bei einer Schlechten Eigenschaft eine Spezielle Erfahrung.
     *
     * @param erfahrung
     *            {@code true}, wenn der Held eine Spezielle Erfahrung auf die
     *            Schlechte Eigenschaft erhalten hat, sonst {@code false}.
     */
    public final void setSpezielleErfahrung( final boolean erfahrung ) {

        if ( erfahrung != spezielleErfahrung ) {

            spezielleErfahrung = erfahrung;

            if ( null != held ) {
                held.notifyObservers( vorNachteil );
            }

            meldeBeobachtern( vorNachteil );
        }
    }

    /**
     * Aendert die Stufe des Wertes.
     *
     * @param neueStufe
     *            des Wertes.
     */
    public final void setStufe( final int neueStufe ) {

        if ( neueStufe != stufe ) {

            stufe = neueStufe;

            if ( null != held ) {

                vorNachteil.stufeGeaendert( held, this );

                held.notifyObservers( vorNachteil );
            }

            meldeBeobachtern( vorNachteil );
        }
    }

    /**
     * Aendert die Stufe des Wertes. Wird benoetigt, um Schlechte Eigenschaften
     * und allgemein Nachteile mit Stufe abzubauen.
     *
     * @param held
     *            der die Stufe des Nachteils aendern will.
     * @param neueStufe
     *            der fuer den Nachteil
     * @param lernmethode
     *            mit der eine Schlechte Eigenschaft evtl. abgebaut werden soll.
     * @return die AP die bezahlt werden muessen (positiver Wert), um die Stufe
     *         zu senken, oder die AP die frei werden (negativer Wert), wenn die
     *         Stufe wieder angehoben wird.
     */
    public final int setStufe(
            final Held held,
            final int neueStufe,
            final Lernmethode lernmethode ) {

        int apKosten = 0;

        if ( neueStufe != stufe ) {

            while ( stufe > neueStufe ) {
                apKosten += senken( held, lernmethode );
            }

            while ( stufe < neueStufe ) {
                apKosten -= steigern();
            }

        }

        return apKosten;
    }

    /**
     * Bestimmt die Variante der Sonderfertigkeit.
     *
     * @see #istVarianteAenderbar()
     * @param dieVariante des Vor-/Nachteils. Das Objekt muss mittels
     *      {@link Object#toString()} sinnvoll in eine String umgewandelt
     *      werden koennen.
     * @throws IllegalArgumentException wenn {@code null} uebergeben wird.
     * @throws UnsupportedOperationException wenn der Vor-/Nachteil keine
     *      Varianten besitzt oder die Variante nicht geaendert werden darf.
     */
    public final void setVariante( final Object dieVariante ) {

        if ( null == dieVariante ) {

            throw new IllegalArgumentException( String.format(
                    "Als Variante fuer den Vor-/Nachteil %1$s wurde null " +
                    "uebergeben. Das ist nicht erlaubt.",
                    getName() ) );
        }

        if ( ! vorNachteil.hatVarianten() ) {

            throw new UnsupportedOperationException( String.format(
                    "Der VorNachteil %1$s besitzt keine Varianten.",
                    vorNachteil.getName() ) );
        }

        if ( ! varianteAenderbar ) {

            throw new UnsupportedOperationException( String.format(
                    "Die Variante von %1$s darf nicht geaendert werden. ",
                    getName() ) );
        }

        final Object alteVariante = variante;

        variante = dieVariante;

        name = String.format( vorNachteil.getNameVariante(), variante );

        if ( null != held ) {

            vorNachteil.geaendertVariante( held, alteVariante, variante );

            held.notifyObservers( vorNachteil );
        }

        meldeBeobachtern( vorNachteil );
    }

    /**
     * @param istVarianteAenderbar
     *            {@code true}, wenn die Variante des Wertes aenderbar sein
     *            soll, sonst {@code false}.
     */
    public final void setVarianteAenderbar( final boolean istVarianteAenderbar ) {
        varianteAenderbar = istVarianteAenderbar;
    }

    /**
     * Hebt die Stufe des Wertes um einen Punkt.
     *
     * @return die AP, die dadurch ggf. frei werden.
     */
    private int steigern() {

        int apKosten = 0;

        if ( stufe < kosten.size() ) {

            apKosten = kosten.get( stufe );
        }

        ++stufe;

        return apKosten;
    }

    @Override
    public final String toString() {
        return name + ( vorNachteil.hatStufen() ? " " + stufe: "" );
    }

}
