/*
    This file is part of the TDE Character Generator (TCG).

    TCG is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TCG is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TCG.  If not, see <http://www.gnu.org/licenses/>.
*/
package de.bernhardjung.dsaprogramm.zauber;

import java.util.ArrayList;
import java.util.List;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.util.FehlendesArgumentException;

/**
 * Repraesentiert den Wert eines Helden.
 *
 * @author Bernhard Jung
 *
 */
public class ZauberWert implements Cloneable {

    /** Die AP die bezahlt wurden um den Wert zu aktivieren. */
    private int aktivierungskosten;

    /** Der Held zu dem der Wert gehoert. */
    private Held held;

    /**
     * Die AP die aufgewendet wurden, um den Zauber zu steigern. An Postion i
     * stehen die Kosten die bezahlt wurden um den Zauber auf Stufe i zu
     * steigern.
     */
    private List< Integer > kosten;

    /**
     * Gibt an ob der Held eine Spezielle Erfahrung auf den Wert erhalten hat.
     */
    private boolean spezielleErfahrung;

    /** Die Stufe die der Held in dem Zauber besitzt. */
    private int stufe;

    /** Die moegliche Variante des Zaubers. */
    private Object variante;

    /** Der Zauber zu dem der Wert gehoert. */
    private final Zauber zauber;

    public ZauberWert( final Zauber derZauber ) {

        this( derZauber, 0 );
    }

    public ZauberWert( final Zauber derZauber, final int dieStufe ) {

        if ( derZauber.hatVarianten() ) {

            throw new FehlendesArgumentException( String.format(
                    "Der Zauber %1$s hat Varianten. Es muss eine Variante " +
                    "angegeben werden.",
                    derZauber.getName() ) );
        }

        aktivierungskosten = 0;

        kosten = new ArrayList< Integer >();

        spezielleErfahrung = false;

        zauber = derZauber;

        variante = "";

        stufe = dieStufe;
    }

    public ZauberWert(
            final Zauber derZauber,
            final Object dieVariante,
            final int dieStufe ) {

        if ( null == dieVariante ) {
            throw new IllegalArgumentException( String.format(
                    "null ist als Argument fuer " +
            		"dieVariante nicht erlaubt.\n" +
            		"derZauber = %1$s\n" +
            		"dieStufe  = %2$s",
            		derZauber,
            		dieStufe ) );
        }

        aktivierungskosten = 0;

        kosten = new ArrayList< Integer >();

        spezielleErfahrung = false;

        zauber = derZauber;

        variante = dieVariante;

        stufe = dieStufe;
    }

    /**
     * @param bonus auf die Stufe des Zaubers.
     */
    final void addiereStufe( final int bonus ) {

        stufe += bonus;

        if ( null != held ) {

            held.notifyObservers( zauber );
        }
    }

    /**
     * @return eine Kopie des Wertes.
     */
    @Override
    public final ZauberWert clone() {

        ZauberWert wert = null ;

        try {

            wert = (ZauberWert) super.clone();

            wert.kosten = new ArrayList< Integer >( kosten );

        } catch (final CloneNotSupportedException e) {
            e.printStackTrace();
        }

        return wert;
    }

    /**
     * Vergleicht zwei Werte nach Zauber und Variante. Die Stufe wird nicht
     * beachtet.
     */
    @Override
    public final boolean equals( final Object obj ) {

        if ( null == obj ) {
            return false;
        }

        final ZauberWert anderer = (ZauberWert) obj;

        return zauber.equals( anderer.zauber )
            && variante.equals( anderer.variante );
    }

    /**
     * @return die AP die der Held bezahlt hat um diesen Wert zu aktivieren.
     */
    public final int getAktivierungskosten() {
        return aktivierungskosten;
    }

    /**
     * @param held
     *            der den Zauber aktivieren will.
     * @return die Aktivierungskosten, die der Held fuer den Zauber zahlen muss.
     */
    public final int getAktivierungskosten( final Held held ) {
        return zauber.getAktivierungskosten( held, this, Lernmethode.KEINE );
    }

    /**
     * @param held
     *            der den Zauber aktivieren will.
     * @param lernmethode
     *            die zum Aktivieren verwendet werden soll.
     * @return die Aktivierungskosten, die der Held fuer den Zauber zahlen muss.
     */
    public final int getAktivierungskosten(
            final Held held,
            final Lernmethode lernmethode ) {

        return zauber.getAktivierungskosten( held, this, lernmethode );
    }

    /**
     * @return Die AP die der Held investiert hat, um den Zauber zu aktivieren
     *         und steigern.
     */
    public final int getApKosten() {

        int apKosten = aktivierungskosten;

        for ( final int ap : kosten ) {

            apKosten += ap;
        }

        return apKosten;
    }

    /**
     * @return den Namen des Zaubers. Er enthaelt den Namen des Zaubers und die
     *         Variante.
     */
    public final String getName() {

        final StringBuilder builder = new StringBuilder();

        builder.append( zauber.getName() );

        if ( zauber.hatVarianten() ) {

            builder.append( " (" );
            builder.append( variante );
            builder.append( ")" );
        }

        return builder.toString();
    }

    /**
     * @return die Variante des Zaubers oder {@code null}, wenn der Zauber
     *         keiner Variante hat.
     */
    public final Object getVariante() {
        return variante;
    }

    /**
     * @param held der den Zauber steigern will.
     * @return den Steigerungsfaktor des Zaubers.
     */
    public final int getSteigerungsfaktor( final Held held ) {
        return zauber.getSteigerungsspalte(
                held, this, stufe, Lernmethode.KEINE );
    }

    /**
     * @param held
     *            der den Zauber steigern will.
     * @param lernmethode
     *            die zum Steigern verwendet werden soll. Wenn der Held eine
     *            Spezielle Erfahrung fuer den Wert besitzt, so wird die hier
     *            uebergebene Lernmethode ignoriert und statt dessen Spezielle
     *            Erfahrung verwendet.
     * @return den Steigerungsfaktor des Zaubers.
     */
    public final int getSteigerungsfaktor(
            final Held held,
            Lernmethode lernmethode ) {

        if ( spezielleErfahrung ) {

            lernmethode = Lernmethode.SPEZIELLE_ERFAHRUNG;
        }

        return zauber.getSteigerungsspalte( held, this, stufe, lernmethode );
    }

    /**
     * Berechnet die Steigerungskosten fuer den Zauber.
     *
     * @param held der den Zauber steigern will.
     * @return die AP die der Held aufbringen muss um den Zauber um einen Punkt
     *         zu steigern.
     */
    public final int getSteigerungskosten( final Held held ) {
        return zauber.getSteigerungskosten(
                held, this, stufe, Lernmethode.KEINE );
    }

    /**
     * Berechnet die Steigerungskosten fuer den Zauber.
     *
     * @param held
     *            der den Zauber steigern will.
     * @param lernmethode
     *            die zum Steigern verwendet werden soll. Wenn der Held eine
     *            Spezielle Erfahrung fuer den Wert besitzt, so wird die hier
     *            uebergebene Lernmethode ignoriert und statt dessen Spezielle
     *            Erfahrung verwendet.
     * @return die AP die der Held aufbringen muss um den Zauber um einen Punkt
     *         zu steigern.
     */
    public final int getSteigerungskosten(
            final Held held,
            Lernmethode lernmethode ) {

        if ( spezielleErfahrung ) {

            lernmethode = Lernmethode.SPEZIELLE_ERFAHRUNG;
        }

        return zauber.getSteigerungskosten( held, this, stufe, lernmethode );
    }

    /**
     * @return die Stufe die der Held in dem Zauber besitzt.
     */
    public final int getStufe() {
        return stufe;
    }

    /**
     * @return den Zauber zu dem der Wert gehoert.
     */
    public final Zauber getZauber() {
        return zauber;
    }

    /**
     * @return {@code true}, wenn der Held eine Spezielle Erfahrung auf den
     *         Wert erhalten hat, sonst {@code false}.
     */
    public final boolean hatSpezielleErfahrung() {
        return spezielleErfahrung;
    }

    /**
     * Ueberprueft ob ein Held die Voraussetzungen fuer den Zauber erfuellt.
     *
     * @param held
     *            der ueberprueft werden soll.
     * @return eine Beschreibung aller aufgetretenen Regelprobleme.
     */
    public final KonfliktBeschreibung pruefen( final Held held ) {
        return zauber.pruefen( held, this );
    }

    /**
     * Senkt die Stufe des Zaubers um einen Punkt.
     * @return die AP die bezahlt wurden um den Zauber auf die bisherige Stufe
     *         zu steigern.
     */
    private final int senken() {
        return kosten.remove( stufe-- );
    }

    public final void setAktivierungskosten( final int ap ) {
        aktivierungskosten = ap;
    }

    /**
     * @param derHeld dem der Wert gehoert.
     */
    public final void setHeld( final Held derHeld ) {
        held = derHeld;
    }

    /**
     * @param apKosten die bezahlt wurden um den Wert auf die aktuelle Stufe zu
     *            steigern.
     */
    private final void setKosten( final int apKosten ) {

        while ( kosten.size() <= stufe ) {

            kosten.add( 0 );
        }

        kosten.set( stufe, apKosten );
    }

    /**
     * Aendert die Stufe des Wertes.
     *
     * @param dieStufe Die neue Stufe des Wertes.
     */
    public final void setStufe( final int dieStufe ) {

        if ( dieStufe != stufe ) {

            stufe = dieStufe;

            if ( null !=  held ) {

                held.notifyObservers( zauber );
            }
        }
    }

    /**
     * @param erfahrung
     *            {@code true}, wenn der Held eine spezielle Erfahrung auf den
     *            Wert erhalten hat, sonst {@code false}.
     */
    public final void setSpezielleErfahrung( final boolean erfahrung ) {

        if ( erfahrung != spezielleErfahrung ) {

            spezielleErfahrung = erfahrung;

            if ( null != held ) {
                held.notifyObservers( zauber );
            }
        }
    }

    /**
     * Andert die Stufe des Zaubers.
     *
     * @param held
     *            dem der Zauber gehoert.
     * @param neueStufe
     *            die der Zauber erhalten soll.
     * @param lernmethode
     *            die verwendet werden soll, falls der Zauber zum Aendern der
     *            Stufe, gesteigert werden muss.
     * @return die AP die zu bezahlen sind, um den Zauber auf die neue Stufe zu
     *         steigern (positiver Wert) oder die AP die frei werden, wenn der
     *         Zauber gesenkt wird (negativer Wert).
     */
    public final int setStufe(
            final Held held,
            final int neueStufe,
            final Lernmethode lernmethode ) {

        int kosten = 0;

        if ( neueStufe != stufe ) {

            while ( stufe < neueStufe ) {
                kosten += steigern( held, lernmethode );
            }

            while ( stufe > neueStufe ) {
                kosten -= senken();
            }

        }

        return kosten;
    }

    /**
     * Aendert die Variante des Wertes.
     *
     * @param neueVariante fuer den Wert.
     * @throws IllegalArgumentException wenn {@code null} uebergeben wird.
     */
    public final void setVariante( final Object neueVariante ) {

        if ( null == neueVariante ) {
            throw new IllegalArgumentException( "null ist als Argument fuer " +
            		"neueVariante nicht erlaubt." );
        }

        if ( ! variante.equals( neueVariante ) ) {

            variante = neueVariante;

            if ( null != held ) {

                held.notifyObservers( zauber );
            }
        }
    }

    /**
     * Hebt die Stufe des Zaubers um einen Punkt. Wenn der Held eine Spezielle
     * Erfahrung fuer den Wert besitzt, so wird die hier uebergebene Lernmethode
     * ignoriert und statt dessen Spezielle Erfahrung verwendet. Nach dem
     * Steigern hat der Wert dann keine Spezielle Erfahrung mehr.
     *
     * @param held
     *            dem der Wert gehoert und der die Stufe anheben will.
     * @param lernmethode
     *            die zum Steigern verwendet werden soll. Wenn der Held eine
     *            Spezielle Erfahrung fuer den Wert besitzt, so wird die hier
     *            uebergebene Lernmethode ignoriert und statt dessen Spezielle
     *            Erfahrung verwendet.
     * @return die AP die bezahlt werden muessen um den Wert anzuheben.
     */
    private final int steigern(
            final Held held,
            Lernmethode lernmethode ) {

        if ( spezielleErfahrung ) {

            lernmethode = Lernmethode.SPEZIELLE_ERFAHRUNG;

            spezielleErfahrung = false;
        }

        final int apKosten;

        if ( held.hatVorgegeben( this ) ) {
            // bis zur Mindeststufe kostest das Steigern nichts.
            final int mindeststufe = held.getMindeststufe( this );

            if ( stufe < mindeststufe ) {

                apKosten = 0;

            } else {

                apKosten = getSteigerungskosten( held, lernmethode );
            }
        } else {

            apKosten = getSteigerungskosten( held, lernmethode );
        }

        ++stufe;

        setKosten( apKosten );

        return apKosten;
    }

    @Override
    public final String toString() {
        return getName() + " " + stufe ;
    }

}
